package dev.fitko.fitconnect.api.services;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentForValidation;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.DecryptionException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;

import java.net.URI;
import java.util.List;
import java.util.Set;
import java.util.UUID;

/**
 * A technical system that accepts submissions on the administration side.
 *
 * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/overview">Receiving Submissions</a>
 */
public interface Subscriber {

    /**
     * Decrypts JWE-encrypted string data.
     *
     * @param privateKey       the private key to decrypt the JWE-encrypted string
     * @param encryptedContent JWE encrypted content that should be decrypted
     * @return the decrypted content as byte[]
     */
    byte[] decryptStringContent(RSAKey privateKey, String encryptedContent) throws DecryptionException;

    /**
     * Polls available {@link SubmissionForPickup}s for a given destinationId.
     *
     * @param destinationId restricts the query to a specific destination
     * @param limit         number of submissions in result (max. is 500)
     * @param offset        position in the dataset
     * @return list of found {@link SubmissionForPickup}s
     */
    Set<SubmissionForPickup> pollAvailableSubmissionsForDestination(UUID destinationId, int offset, int limit) throws RestApiException;

    /**
     * Polls available {@link SubmissionForPickup}s for all destinations the {@link Subscriber} is assigned to.
     *
     * @param limit  number of submissions in result (max. is 500)
     * @param offset position in the dataset
     * @return list of found {@link SubmissionForPickup}s
     */
    Set<SubmissionForPickup> pollAvailableSubmissions(int offset, int limit);

    /**
     * Gets a {@link Submission}.
     *
     * @param submissionId the unique identifier of a {@link Submission}
     * @return the requested {@link Submission}
     */
    Submission getSubmission(UUID submissionId) throws RestApiException;

    /**
     * Loads encrypted {@link ApiAttachment} for a {@link Submission}.
     *
     * @param submissionId the unique identifier of a {@link Submission}
     * @param attachmentId unique identifier of the attachments
     * @return encrypted JWE string of attachment
     */
    String fetchAttachment(UUID submissionId, UUID attachmentId) throws RestApiException;

    /**
     * Retrieve the entire event log for a submissions caseId of a specific destination.
     *
     * @param caseId        unique identifier of the case the log should be retrieved for
     * @param destinationId unique identifier of the {@link Destination} the log should be retrieved for
     * @return List of {@link EventLogEntry}s for the given case
     */
    List<EventLogEntry> getEventLog(UUID caseId, UUID destinationId) throws RestApiException;

    /**
     * Validates the {@link Metadata} structure and contents to ensure its correctness.
     *
     * @param metadata the {@link Metadata} object that is validated
     * @param submission the {@link Submission} of the validated metadata
     * @param authenticationTags the {@link AuthenticationTags} of the validated metadata
     * @return a {@link ValidationResult}, contains an error if the {@link Metadata} is invalid, e.g. doesn't match the schema
     */
    ValidationResult validateMetadata(Metadata metadata, Submission submission, AuthenticationTags authenticationTags);

    /**
     * Validates the attachment structure and contents to ensure its correctness.
     *
     * @param attachmentsForValidation list of attachments containing the hash, decrypted and encrypted data needed for validation
     * @param authenticationTags the {@link AuthenticationTags} of the validated attachments

     * @return a {@link ValidationResult}, contains an error if the attachment is invalid
     */
    ValidationResult validateAttachments(List<AttachmentForValidation> attachmentsForValidation, AuthenticationTags authenticationTags);

    /**
     * Validates the submission data structure against a given JSON-schema to ensure its correctness.
     *
     * @param submissionData the submission data object that is validated
     * @param schemaUri schema to validate against
     * @return a {@link ValidationResult}, contains an error if the submission data is invalid or doesn't match the schema
     */
    ValidationResult validateSubmissionDataSchema(String submissionData, URI schemaUri);

    /**
     * Validates the {@link Data} structure and contents to ensure its correctness.
     *
     * @param data               the unencrypted data as byte[]
     * @param submission         the {@link Submission} of the validated data
     * @param metadata           the {@link Metadata} of the validated data
     * @param authenticationTags the {@link AuthenticationTags} of the validated data
     * @return a {@link ValidationResult}, contains an error if the {@link Metadata} is invalid, e.g. doesn't match the schema
     */
    ValidationResult validateData(byte[] data, Submission submission, Metadata metadata, AuthenticationTags authenticationTags);

    /**
     * Validates data integrity of {@link ApiAttachment} or {@link Data}.
     *
     * @param originalHash original hash value of raw data as hex string
     * @param data         unencrypted data as byte[]
     * @return a {@link ValidationResult}, contains an error if the hash values are not equal
     */
    ValidationResult validateHashIntegrity(String originalHash, byte[] data);

    /**
     * Checks if a received callback can be trusted by validating the provided request data
     *
     * @param hmac           authentication code provided by the callback
     * @param timestamp      timestamp provided by the callback
     * @param httpBody       HTTP body provided by the callback
     * @param callbackSecret secret owned by the client, which is used to calculate the hmac
     * @return {@code true} if hmac and timestamp provided by the callback meet the required conditions
     */
    ValidationResult validateCallback(String hmac, Long timestamp, String httpBody, String callbackSecret);

    /**
     * Sends a confirmation event if the received submission matches all validations.
     *
     * @param eventPayload contains submissionId, caseId, destination and a list of problems
     * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/process-and-acknowledge">Process And Acknowledge</a>
     */
    void acceptSubmission(EventPayload eventPayload);

    /**
     * Sends a rejection event if the received submission violates any validation rule.
     *
     * @param eventPayload contains submissionId, caseId, destination and a list of problems
     * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/process-and-acknowledge">Process And Acknowledge</a>
     */
    void rejectSubmission(EventPayload eventPayload);

    /**
     * Get event from event-log filtered by a given event and submission.
     *
     * @param event event type to filter the event-log by
     * @param submission submission to filter the event-log by
     * @return {@link AuthenticationTags} of metadata, data and attachments
     */
    AuthenticationTags getAuthenticationTagsForEvent(Event event, Submission submission);

    /**
     * Retrieve the current status of a {@link Submission}.
     *
     * @param destinationId unique identifier of the {@link Destination} the status should be retrieved for
     * @param caseId unique identifier of the case the status should be retrieved for
     * @param submissionId unique identifier of the submission the status should be retrieved for
     *
     * @return {@link SubmissionStatus} the current status
     */
    SubmissionStatus getLastedEvent(final UUID destinationId, final UUID caseId, final UUID submissionId) throws RestApiException;
}
