package dev.fitko.fitconnect.core.http;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.github.tomakehurst.wiremock.WireMockServer;
import com.github.tomakehurst.wiremock.client.WireMock;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import okhttp3.RequestBody;
import okhttp3.Response;
import okhttp3.ResponseBody;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.util.Map;
import java.util.Objects;

import static com.github.tomakehurst.wiremock.client.WireMock.equalTo;
import static com.github.tomakehurst.wiremock.client.WireMock.matching;
import static com.github.tomakehurst.wiremock.client.WireMock.ok;
import static com.github.tomakehurst.wiremock.client.WireMock.urlEqualTo;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.nullValue;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class HttpClientTest {

    private final WireMockServer wireMockServer = new WireMockServer();

    @BeforeEach
    public void startWireMock() {
        this.wireMockServer.start();
    }

    @AfterEach
    public void shutDownWireMock() {
        this.wireMockServer.stop();
    }

    @Test
    public void createRequestBodyWithImplicitContentType() throws JsonProcessingException {

        RequestBody requestBody = new HttpClient(true).createRequestBody(Map.of(), "abc");

        assertThat(Objects.requireNonNull(requestBody.contentType()).toString(), is(MimeTypes.APPLICATION_JSON));
    }

    /**
     * See comments at {@link HttpClient#createRequestBody(Map, Object)} for more information
     */
    @Test
    public void createRequestBodyWithNoColonsInString() throws IOException {

        RequestBody requestBody = new HttpClient(true).createRequestBody(Map.of(), "abc");

        assertThat(requestBody.contentLength(), is(3L));
    }

    @Test
    public void createRequestBodyWithExplicitContentType() throws IOException {

        RequestBody requestBody = new HttpClient(true).createRequestBody(
                Map.of("Content-Type", "application/json"), 10);

        assertThat(requestBody.contentLength(), is(2L));
        assertThat(Objects.requireNonNull(requestBody.contentType()).toString(), is("application/json"));
    }

    @Test
    public void respondWithRestApiExceptionIfConfigured() {

        Response response = mock(Response.class);
        when(response.code()).thenReturn(400);
        when(response.isSuccessful()).thenReturn(false);

        RestApiException exception = Assertions.assertThrows(RestApiException.class,
                () -> new HttpClient(true).evaluateStatusAndRespond(response, String.class));

        assertThat(exception.getMessage(), is("HTTP request failed."));
        assertThat(exception.getStatusCode(), is(400));
    }

    @Test
    public void respondWithNoExceptionIfNotConfigured() throws IOException {

        Response response = mock(Response.class);
        when(response.code()).thenReturn(400);
        when(response.isSuccessful()).thenReturn(false);
        when(response.body()).thenReturn(mock(ResponseBody.class));

        HttpResponse<?> httpResponse = new HttpClient(false).evaluateStatusAndRespond(response, String.class);

        assertThat(httpResponse.getStatusCode(), is(400));
    }

    @Test
    public void respondWithIOExceptionIfResponseBodyIsNull() {

        Response response = mock(Response.class);
        when(response.isSuccessful()).thenReturn(true);

        IOException exception = Assertions.assertThrows(IOException.class,
                () -> new HttpClient(true).evaluateStatusAndRespond(response, String.class));

        assertThat(exception.getMessage(), is("Response body is null."));
    }

    @Test
    public void respondWithEmptyBodyIfResponseTypeIsVoid() throws IOException {

        Response response = mock(Response.class);
        when(response.code()).thenReturn(200);
        when(response.isSuccessful()).thenReturn(true);
        when(response.body()).thenReturn(mock(ResponseBody.class));

        HttpResponse<?> httpResponse = new HttpClient(true).evaluateStatusAndRespond(response, Void.class);

        assertThat(httpResponse.getBody(), nullValue());
        assertThat(httpResponse.getStatusCode(), is(200));
    }

    @Test
    public void respondWithRawStringIfResponseTypeIsVoid() throws IOException {

        ResponseBody responseBody = mock(ResponseBody.class);
        when(responseBody.string()).thenReturn("abc");
        Response response = mock(Response.class);
        when(response.code()).thenReturn(200);
        when(response.isSuccessful()).thenReturn(true);
        when(response.body()).thenReturn(responseBody);

        HttpResponse<?> httpResponse = new HttpClient(true).evaluateStatusAndRespond(response, String.class);

        assertThat(httpResponse.getBody(), is("abc"));
        assertThat(httpResponse.getStatusCode(), is(200));
    }

    @Test
    public void respondWithComplexBodyIfResponseTypeComplex() throws IOException {

        ResponseBody responseBody = mock(ResponseBody.class);
        when(responseBody.string()).thenReturn("{ \"property\": \"value\" }");
        Response response = mock(Response.class);
        when(response.code()).thenReturn(200);
        when(response.isSuccessful()).thenReturn(true);
        when(response.body()).thenReturn(responseBody);

        HttpResponse<?> httpResponse = new HttpClient(true).evaluateStatusAndRespond(response, Map.class);

        assertThat(httpResponse.getBody(), is(Map.of("property", "value")));
        assertThat(httpResponse.getStatusCode(), is(200));
    }

    @Test
    public void get() throws IOException {

        WireMock.configureFor("localhost", 8080);
        this.wireMockServer.stubFor(WireMock.get(urlEqualTo("/test"))
                .withHeader("User-Agent", equalTo("online service"))
                .willReturn(ok().withBody("success"))
        );

        HttpResponse<String> httpResponse = new HttpClient(true)
                .get("http://localhost:8080/test", Map.of("User-Agent", "online service"), String.class);

        assertThat(httpResponse.getStatusCode(), is(200));
        assertThat(httpResponse.getBody(), is("success"));
    }

    @Test
    public void post() throws IOException {

        WireMock.configureFor("localhost", 8080);
        this.wireMockServer.stubFor(WireMock.post(urlEqualTo("/test"))
                .withHeader("User-Agent", equalTo("online service"))
                .withRequestBody(matching("data"))
                .willReturn(ok().withBody("success"))
        );

        HttpResponse<String> httpResponse = new HttpClient(true)
                .post("http://localhost:8080/test", Map.of("User-Agent", "online service"), "data", String.class);

        assertThat(httpResponse.getStatusCode(), is(200));
        assertThat(httpResponse.getBody(), is("success"));
    }

    @Test
    public void put() throws IOException {

        WireMock.configureFor("localhost", 8080);
        this.wireMockServer.stubFor(WireMock.put(urlEqualTo("/test"))
                .withHeader("User-Agent", equalTo("online service"))
                .withRequestBody(matching("data"))
                .willReturn(ok().withBody("success"))
        );

        HttpResponse<String> httpResponse = new HttpClient(true)
                .put("http://localhost:8080/test", Map.of("User-Agent", "online service"), "data", String.class);

        assertThat(httpResponse.getStatusCode(), is(200));
        assertThat(httpResponse.getBody(), is("success"));
    }
}
